#include <avr/io.h>
#include <util/delay.h>

#define DHTPIN 6       // Pin connected to DHT sensor
#define BUZZER_PIN 9   // Pin connected to the buzzer
#define LCD_ADDR 0x3F  // I2C address of the LCD
#define SD_CS_PIN 10   // Pin connected to the SD card CS

#define MOSI_PIN 11
#define MISO_PIN 12
#define SCK_PIN 13
#define SS_PIN 10
#define SDA_PIN A4     // I2C data pin
#define SCL_PIN A5     // I2C clock pin

// LCD Commands and Bits
#define LCD_BACKLIGHT   0x08
#define LCD_NOBACKLIGHT 0x00
#define En 0b00000100  // Enable bit
#define Rw 0b00000010  // Read/Write bit
#define Rs 0b00000001  // Register select bit

#define BLOCK_SIZE 512

unsigned long previousMillis = 0; 
const long interval = 1000; 

void setup() {
  pinMode(BUZZER_PIN, OUTPUT); 
  Serial.begin(9600);
  
  // Initialize LCD
  lcdInit();

  // Initialize SD card
  Serial.print("Initializing SD card...");
  if (!sdInit()) {
    Serial.println("initialization failed!");
    while (1);
  }
  Serial.println("initialization done.");
}

void loop() {
  
  unsigned long currentMillis = millis();
  
  if (currentMillis - previousMillis >= interval) {
    
    previousMillis = currentMillis;
    
    float t, h;
    if (!readDHT(t, h)) {
      Serial.println("Failed to read from DHT sensor!");
      return;
    }

    if (t > 27) {
      tone(BUZZER_PIN, 1000); 
    } else {
      noTone(BUZZER_PIN); 
    }
    
    Serial.print("Temperature: "); 
    Serial.print(t);
    Serial.println(" *C ");
    Serial.print("Humidity: "); 
    Serial.print(h);
    Serial.println(" %");
    
    lcdClear(); 
    lcdSetCursor(0, 0); 
    lcdPrint("Temp: ");
    lcdPrint(t);
    lcdPrint(" C");


    if (t > 27) {
      lcdClear(); 
      lcdSetCursor(0, 0); 
      lcdPrint("INCENDIU");
    }

    if (sdWriteData(currentMillis, t, h)) {
      Serial.println("Data written to SD card.");
    } else {
      Serial.println("Error writing to SD card.");
    }
  }
}


bool readDHT(float &temperature, float &humidity) {
  uint8_t data[5] = {0};

  pinMode(DHTPIN, OUTPUT);
  digitalWrite(DHTPIN, LOW);
  delay(20);
  digitalWrite(DHTPIN, HIGH);
  delayMicroseconds(40);
  pinMode(DHTPIN, INPUT_PULLUP);

  if (!waitForSignal(LOW, 80) || !waitForSignal(HIGH, 80) || !waitForSignal(LOW, 80)) {
    return false;
  }

  for (uint8_t i = 0; i < 5; i++) {
    data[i] = readByte();
  }

  if (data[4] == ((data[0] + data[1] + data[2] + data[3]) & 0xFF)) {
    humidity = ((data[0] << 8) + data[1]) * 0.1;
    temperature = (((data[2] & 0x7F) << 8) + data[3]) * 0.1;
    if (data[2] & 0x80) {
      temperature = -temperature;
    }
    return true;
  }
  return false;
}

bool waitForSignal(bool level, unsigned long timeout) {
  unsigned long startTime = micros();
  while (digitalRead(DHTPIN) != level) {
    if (micros() - startTime > timeout) {
      return false;
    }
  }
  return true;
}

uint8_t readByte() {
  uint8_t result = 0;
  for (uint8_t i = 0; i < 8; i++) {
    if (!waitForSignal(HIGH, 50)) {
      return result;
    }
    delayMicroseconds(28); 
    result <<= 1;
    if (digitalRead(DHTPIN) == HIGH) {
      result |= 1;
    }
    if (!waitForSignal(LOW, 50)) {
      return result;
    }
  }
  return result;
}


void i2cInit() {
  pinMode(SDA_PIN, OUTPUT);
  pinMode(SCL_PIN, OUTPUT);
}

void i2cStart() {
  digitalWrite(SDA_PIN, LOW);
  delayMicroseconds(4);
  digitalWrite(SCL_PIN, LOW);
}

void i2cStop() {
  digitalWrite(SDA_PIN, LOW);
  delayMicroseconds(4);
  digitalWrite(SCL_PIN, HIGH);
  delayMicroseconds(4);
  digitalWrite(SDA_PIN, HIGH);
}

void i2cWriteBit(bool bit) {
  digitalWrite(SDA_PIN, bit);
  delayMicroseconds(4);
  digitalWrite(SCL_PIN, HIGH);
  delayMicroseconds(4);
  digitalWrite(SCL_PIN, LOW);
}

bool i2cReadBit() {
  pinMode(SDA_PIN, INPUT);
  digitalWrite(SCL_PIN, HIGH);
  delayMicroseconds(4);
  bool bit = digitalRead(SDA_PIN);
  digitalWrite(SCL_PIN, LOW);
  pinMode(SDA_PIN, OUTPUT);
  return bit;
}

void i2cWriteByte(uint8_t byte) {
  for (uint8_t bit = 0; bit < 8; bit++) {
    i2cWriteBit(byte & 0x80);
    byte <<= 1;
  }
  i2cReadBit(); // Read ACK bit
}

uint8_t i2cReadByte(bool ack) {
  uint8_t byte = 0;
  pinMode(SDA_PIN, INPUT);
  for (uint8_t bit = 0; bit < 8; bit++) {
    digitalWrite(SCL_PIN, HIGH);
    delayMicroseconds(4);
    byte = (byte << 1) | digitalRead(SDA_PIN);
    digitalWrite(SCL_PIN, LOW);
    delayMicroseconds(4);
  }
  pinMode(SDA_PIN, OUTPUT);
  i2cWriteBit(!ack);
  return byte;
}

// LCD Functions
void lcdInit() {
  i2cInit();
  
  lcdSendCommand(0x03);
  lcdSendCommand(0x03);
  lcdSendCommand(0x03);
  lcdSendCommand(0x02);

  lcdSendCommand(0x28); // Function set: 4-bit mode, 2 lines, 5x8 dots
  lcdSendCommand(0x0C); // Display control: display on, cursor off, blink off
  lcdSendCommand(0x06); // Entry mode set: increment automatically, no display shift
  lcdSendCommand(0x01); // Clear display
  delay(2);
}

void lcdSendCommand(uint8_t command) {
  lcdSend(command, 0);
}

void lcdSendData(uint8_t data) {
  lcdSend(data, Rs);
}

void lcdSend(uint8_t value, uint8_t mode) {
  uint8_t highNibble = value & 0xf0;
  uint8_t lowNibble = (value << 4) & 0xf0;
  lcdWrite4bits(highNibble | mode);
  lcdWrite4bits(lowNibble | mode);
}

void lcdWrite4bits(uint8_t value) {
  i2cStart();
  i2cWriteByte(LCD_ADDR << 1);
  i2cWriteByte(value | En | LCD_BACKLIGHT);
  i2cWriteByte((value & ~En) | LCD_BACKLIGHT);
  i2cStop();
  delayMicroseconds(50);
}

void lcdClear() {
  lcdSendCommand(0x01); // Clear display
  delay(2); // Delay for clear display command
}

void lcdSetCursor(uint8_t col, uint8_t row) {
  static const uint8_t rowOffsets[] = { 0x00, 0x40, 0x14, 0x54 };
  lcdSendCommand(0x80 | (col + rowOffsets[row]));
}

void lcdPrint(const char* str) {
  while (*str) {
    lcdSendData(*str++);
  }
}

void lcdPrint(float num) {
  char buf[10];
  dtostrf(num, 6, 2, buf);
  lcdPrint(buf);
}

// SD Card Functions
void spiBegin() {
  pinMode(MOSI_PIN, OUTPUT);
  pinMode(MISO_PIN, INPUT);
  pinMode(SCK_PIN, OUTPUT);
  pinMode(SS_PIN, OUTPUT);
  
  // Enable SPI, Master, set clock rate fck/128
  SPCR = (1 << SPE) | (1 << MSTR) | (1 << SPR1) | (1 << SPR0);
  SPSR &= ~(1 << SPI2X); // Clear the SPI2X bit
}

void spiEnd() {
  SPCR &= ~(1 << SPE);
}

uint8_t spiTransfer(uint8_t data) {
  SPDR = data;
  while (!(SPSR & (1 << SPIF)));
  return SPDR;
}

void spiSelect() {
  digitalWrite(SD_CS_PIN, LOW);
}

void spiDeselect() {
  digitalWrite(SD_CS_PIN, HIGH);
}

bool sdInit() {
  spiBegin();
  spiDeselect();
  for (uint8_t i = 0; i < 10; i++) {
    spiTransfer(0xFF); // Send 80 clock pulses
  }
  spiSelect();

  // CMD0: Go to idle state
  if (sdCommand(0x40, 0x00000000) != 0x01) {
    Serial.println("CMD0 failed");
    return false;
  }

  // CMD8: Check voltage range and version
  if (sdCommand(0x48, 0x000001AA) != 0x01) {
    Serial.println("CMD8 failed");
    return false;
  }

  // ACMD41: Initiate initialization process
  unsigned long startTime = millis();
  while (sdACmd(0x69, 0x40000000) != 0x00) {
    if (millis() - startTime > 1000) { // Wait up to 1 second
      Serial.println("ACMD41 timeout");
      return false;
    }
    delay(10); // Small delay before retrying
  }

  spiDeselect();
  return true;
}


uint8_t sdCommand(uint8_t cmd, uint32_t arg) {
  spiTransfer(cmd | 0x40);
  spiTransfer(arg >> 24);
  spiTransfer(arg >> 16);
  spiTransfer(arg >> 8);
  spiTransfer(arg);
  uint8_t crc = 0xFF;
  if (cmd == 0x40) crc = 0x95; // CRC for CMD0
  if (cmd == 0x48) crc = 0x87; // CRC for CMD8
  spiTransfer(crc);
  for (uint8_t i = 0; i < 8; i++) {
    uint8_t response = spiTransfer(0xFF);
    if (response != 0xFF) {
      return response;
    }
  }
  return 0xFF;
}

uint8_t sdACmd(uint8_t cmd, uint32_t arg) {
  if (sdCommand(0x77, 0x00000000) > 0x01) { // CMD55 must return 0x01 (R1_IDLE_STATE)
    return 0xFF;
  }
  return sdCommand(cmd, arg);
}


bool sdWriteData(unsigned long time, float temperature, float humidity) {
  char data[128];
  snprintf(data, sizeof(data), "Time: %lu s, Temp: %.2f C, Hum: %.2f %%\n", time / 1000, temperature, humidity);
  return sdWriteBlock(data);
}

bool sdWriteBlock(const char *data) {
  spiSelect();
  if (sdCommand(0x58, 0x00000000) != 0x00) {
    spiDeselect();
    return false;
  }
  spiTransfer(0xFE); // Start token
  for (uint16_t i = 0; i < BLOCK_SIZE; i++) {
    spiTransfer(data[i]);
  }
  spiTransfer(0xFF); // Dummy CRC
  spiTransfer(0xFF);
  uint8_t response = spiTransfer(0xFF);
  spiDeselect();
  return (response & 0x1F) == 0x05;
}
